using System;

namespace MQ_Postcard {
	/// <summary>
	/// Postcard is a value object to hold the postcard data.
	/// It is used for both incoming and outgoing cards.
	/// </summary>
	public class Postcard {
		private string _sender;
		public  string sender {    get {return _sender;}    set {_sender= value;}  }

		private string _addressee;
		public  string addressee {    get {return _addressee;}    set {_addressee= value;}  }

		private string _rtnMachine;
		public  string rtnMachine {    get {return _rtnMachine;}    set {_rtnMachine= value;}  }

		private System.DateTime _posted;
		public  System.DateTime posted {    get {return _posted;}    set {_posted= value;}  }

		private string _message;
		public  string message {    get {return _message;}    set {_message= value;}  }

		private int _status;
		public  int status {    get {return _status;}    set {_status= value;}  }

		private string _direction;
		public string direction { get { return _direction;} set {_direction=value;} } 

		public string Name {
			get {
				if (direction=="sent") return addressee;
				else if (direction=="received") return sender;
				else return "";
			}
		}

		public Postcard() {
			status=2; //not initialized
			message="(not initialized)";
			sender="--";
			addressee="--";
			rtnMachine= "localhost";
			posted= System.DateTime.Now;
		}
		public Postcard(String s, string a, string m) {
			sender=s;
			message=m;
			addressee=a;
			posted= System.DateTime.Now;
			status=0;
			rtnMachine="localhost";
		}
	}

	/// <summary>
	/// The PostOffice object does 2 things: sends postcards and gets postcards.
	/// </summary>
	public class PostOffice : System.IDisposable {
		private IBM.WMQ.MQQueueManager _mqm;
		private string _queueName= "postcard";
		private string _QueueMgrName= null;
		private IBM.WMQ.MQQueue _q;
		private IBM.WMQ.MQPutMessageOptions _pmo= new IBM.WMQ.MQPutMessageOptions ();

		const int PART1_LENGTH= 512;
		const int PART2_LENGTH= 28;
		const int PART3_LENGTH= 50;
		const int PART4_LENGTH= 512;

		private const int MAX_BROWSE_FAILURES_BEFORE_RESET= 3;
		private const int WAIT_ON_GET= 20; // milliseconds
		private const int WAIT_ON_BROWSE= 50; // milliseconds (if I make this high, sending takes a long time)

		public PostOffice(string qmname) {
			if ((qmname=="(Default Queue Manager)") || (qmname=="(Default)")) {
				_mqm = new IBM.WMQ.MQQueueManager(); 
				_QueueMgrName= null;
			}
			else {
				_mqm = new IBM.WMQ.MQQueueManager(qmname);
				_QueueMgrName= qmname;
			}
			_pmo.Options |= IBM.WMQ.MQC.MQPMO_NONE ;
		}
		
		~PostOffice() {
			Dispose(false);
		}

		public void Dispose() {
			Dispose(true);
		}
		bool hasBeenDisposed= false;

		protected void Dispose(bool disposing_and_not_finalizing_yet) {
			if (!hasBeenDisposed) {
				lock(this) {
					hasBeenDisposed=true;
					if(disposing_and_not_finalizing_yet) {
						System.GC.SuppressFinalize(this);
						// Dispose of managed objects here, if disposing.
					}
					// free external resources here
					if (_q!=null) _q.Close();
					if (_mqm!=null) 	_mqm.Disconnect();
				}
			}
		}

		public string QueueManagerName {
			get { 
				if (mqm==null) return null;
				return mqm.Name; 
			}
		}

		public String QueueName {
			get { return _queueName; }
			set {
				if ((value != _queueName) || (_q == null)) {
					_queueName= value;
					int openOptions= 
						IBM.WMQ.MQC.MQOO_INPUT_AS_Q_DEF +
						IBM.WMQ.MQC.MQOO_OUTPUT + 
						//IBM.WMQ.MQC.MQOO_INPUT_SHARED + 
						IBM.WMQ.MQC.MQOO_FAIL_IF_QUIESCING;
					_q= mqm.AccessQueue(_queueName,openOptions);
				}
			}
		}

		public IBM.WMQ.MQQueueManager mqm {
			get { return _mqm; }
		}

		private byte[] getBuffer= new byte[1700];

		public Postcard GetPostcard(string messageRecipient) {
			Postcard p= new Postcard();
			p.addressee= messageRecipient;
			try {
				bool stop= false; 
				int tries=0; 
				string coId;
				while (!stop) {
					try {
						IBM.WMQ.MQMessage msg= new IBM.WMQ.MQMessage();
						IBM.WMQ.MQGetMessageOptions getOptions= new IBM.WMQ.MQGetMessageOptions();
						getOptions.Options |= IBM.WMQ.MQC.MQGMO_WAIT;
						getOptions.WaitInterval= WAIT_ON_GET;  // in milliseconds, or MQWI_UNLIMITED 

						// some MQ Postcard tools use the Recipient name as the Correlation ID,
						// and a well-known message ID; some of them (Java version) use no Message ID, 
						// and the CoID is %MQPC% + the name.    We try both.  
						if (tries==0) {
							coId= messageRecipient;
							msg.MessageId= MQUtils.MqPostcardMessageId(); // byte[]
						}
						else {
							coId= "%MQPC%" + messageRecipient;
						}
						msg.CorrelationId= MQUtils.StringToCorrelationId(coId);
						_q.Get(msg, getOptions);
						stop=true; 
						if (msg.Format == IBM.WMQ.MQC.MQFMT_STRING) {
							getBuffer= msg.ReadBytes(getBuffer.Length);
							p.message= MQUtils.StringFromBuffer(getBuffer,0,PART1_LENGTH/2);
							p.sender= MQUtils.StringFromBuffer(getBuffer,PART1_LENGTH+PART2_LENGTH, PART3_LENGTH/2);
							p.rtnMachine= MQUtils.StringFromBuffer(getBuffer,PART1_LENGTH+PART2_LENGTH+ PART3_LENGTH, PART4_LENGTH/2);
							p.posted= MQUtils.GetDateTime(getBuffer, 524);
							p.direction= "received";
							p.status= 0; 
						}
						else {
							p.status=3;
							p.message= "Unknown message format";
						}
					}
					catch (IBM.WMQ.MQException mqex) {
						tries++;
						if (tries >1) stop=true;
						if (mqex.Reason == IBM.WMQ.MQC.MQRC_NO_MSG_AVAILABLE) {
							p.message= "(No message available)";
							p.status=1;  // no message
						}
						else {
							p.message= "Error Getting: " + MQUtils.MQExceptionToString(mqex);
						}
					}
				}

			}
			catch (IBM.WMQ.MQException mqex) {
				p.message= "Error Connecting: " + MQUtils.MQExceptionToString(mqex);
			}
			return p;
		}
	
		private byte[] sendBuffer= new byte[1700]; //fixed size for postcard app

		private String _lastSender= null; 
		private String _lastRtnMachine= null; 
		private String _lastMsgToSend= null; 

		public Postcard SendPostcard( Postcard p) {
			p.rtnMachine= System.Net.Dns.GetHostName(); // ??

			try {
				string realMessageToSend = p.sender + " says to " + p.addressee + ": " + p.message;
				if (_lastMsgToSend!=realMessageToSend) {
					MQUtils.FillByteArray(ref sendBuffer, realMessageToSend,0,PART1_LENGTH);
					_lastMsgToSend= realMessageToSend;
				}
				if (_lastSender!=p.sender) {
					MQUtils.FillByteArray(ref sendBuffer, p.sender,PART1_LENGTH + PART2_LENGTH, PART3_LENGTH );
					_lastSender= p.sender;
				}
				if (_lastRtnMachine!=p.rtnMachine) {
					MQUtils.FillByteArray(ref sendBuffer, p.rtnMachine,PART1_LENGTH + PART2_LENGTH + PART3_LENGTH, PART4_LENGTH);
					_lastRtnMachine= p.rtnMachine;
				}
				sendBuffer[516]=1;
				sendBuffer[520]=0;
				// always update the time:
				MQUtils.FillDateTime(ref sendBuffer,524);
				IBM.WMQ.MQMessage msg= new IBM.WMQ.MQMessage();
				msg.Write(sendBuffer);  // buffer.Length
					
				msg.CorrelationId= MQUtils.StringToCorrelationId(p.addressee);
				msg.MessageId= MQUtils.MqPostcardMessageId();
				msg.CharacterSet= 1200;
				msg.Persistence= IBM.WMQ.MQC.MQPER_PERSISTENT; // in case the queue is not peristent
				msg.Format= IBM.WMQ.MQC.MQFMT_STRING;
				msg.ApplicationOriginData= ".NET Postcard sample!";
				msg.Expiry= 86400*10 ; // An expiry time expressed in tenths of a second
				//pmo.Options += IBM.WMQ.MQC.MQPMO_SYNCPOINT;
				_q.Put(msg, _pmo);
				p.direction= "sent";
			}
			catch (IBM.WMQ.MQException mqex1) {
				throw new Exception("Cannot Send Postcard: " + MQUtils.MQExceptionToString(mqex1), mqex1);
			}
			return p;
		}

		private System.Windows.Forms.Form _formToSignal;
		private System.Windows.Forms.MethodInvoker _mi;
		private System.Threading.Thread QueueWatcher= null;
		private System.Threading.AutoResetEvent QueueBrowserFinished;
		private String _recipientToWatchFor;

		public void StartWatching(String recipientName, System.Windows.Forms.Form f, System.Windows.Forms.MethodInvoker mi) {
			StopWatching();			
			_formToSignal= f;
			_mi= mi;
			_recipientToWatchFor= recipientName; 
			InitializeQueueWatcher(); 
		}

	
		public void StopWatching() {
			if (QueueWatcher != null) {
				QueueWatcher.Interrupt();
				QueueWatcher= null;
				if (QueueBrowserFinished!= null) {
					QueueBrowserFinished.WaitOne();
					QueueBrowserFinished= null;
				}
			}
		}

		private void InitializeQueueWatcher() {
			QueueBrowserFinished = new System.Threading.AutoResetEvent(false);
			QueueWatcher = new System.Threading.Thread(new System.Threading.ThreadStart(WatchQueue));
			QueueWatcher.IsBackground = true;
			QueueWatcher.Start();
		}

		private void WatchQueue() {
			bool cease= false;
			while (!cease) {
				try {
					// The watcher thread needs its own MQQueueManager and MQQueue instances
					IBM.WMQ.MQQueueManager 
						WatcherThreadQm = (_QueueMgrName==null)  ? 
						new IBM.WMQ.MQQueueManager() : 
						new IBM.WMQ.MQQueueManager(_QueueMgrName);

					IBM.WMQ.MQQueue q= WatcherThreadQm.AccessQueue(QueueName,openOptionsForWatcher);

					bool reset= false; 
					int MsgsGotConsecutively = 0;
					int failures=0; 
					int tries=0; 
					string coId;

					IBM.WMQ.MQGetMessageOptions getOptions= new IBM.WMQ.MQGetMessageOptions();
					getOptions.Options |= IBM.WMQ.MQC.MQGMO_WAIT | IBM.WMQ.MQC.MQGMO_BROWSE_FIRST; 
					getOptions.WaitInterval= WAIT_ON_BROWSE;  // in milliseconds, or MQWI_UNLIMITED 

					// The WaitInterval should be brief.  There is a lock somewhere in MQ.  When
					// this app calls mq.Put and mq.Get concurrently, the Put can wait for the lock
					// held by mq.Get.  If the WaitInterval is brief, messages can be sent more quickly.
					while (!reset) {
						try {
							IBM.WMQ.MQMessage msg= new IBM.WMQ.MQMessage();

							// some MQ Postcard tools use the Recipient name as the Correlation ID,
							// and a well-known message ID; some of them  (eg Java) use no Message ID, 
							// and the CoID is %MQPC% + the name.    We try both.  
							if (tries==0) {
								coId= _recipientToWatchFor;
								msg.MessageId= MQUtils.MqPostcardMessageId();
							}
							else {
								coId= "%MQPC%" + _recipientToWatchFor;
							}
							msg.CorrelationId= MQUtils.StringToCorrelationId(coId);
							tries++;
							q.Get(msg, getOptions);
							MsgsGotConsecutively++;
							if (msg.Format == IBM.WMQ.MQC.MQFMT_STRING) {
								// use BeginInvoke to marshall across threads (forms are single-threaded). 
								_formToSignal.BeginInvoke(_mi);
								System.Threading.Thread.Sleep(WAIT_ON_BROWSE);
							}
						}
						catch (IBM.WMQ.MQException mqex) {
							failures++;
							reset= (tries >= 2) ;
							if (mqex.Reason != IBM.WMQ.MQC.MQRC_NO_MSG_AVAILABLE) {
								System.Console.WriteLine("Error Getting: " + MQUtils.MQExceptionToString(mqex)); 
							}
						}
					}
				}
				catch (IBM.WMQ.MQException mqex) {
					System.Console.WriteLine("Error Connecting: " + MQUtils.MQExceptionToString(mqex)); 
				}
				catch (System.Threading.ThreadInterruptedException ex1) {
					cease= true; 
				}
			}

			QueueBrowserFinished.Set();
		
		}


		private const int openOptionsForWatcher= 
			IBM.WMQ.MQC.MQOO_INPUT_AS_Q_DEF +
			IBM.WMQ.MQC.MQOO_OUTPUT + 
			IBM.WMQ.MQC.MQOO_BROWSE +
			//IBM.WMQ.MQC.MQOO_INPUT_SHARED + 
			IBM.WMQ.MQC.MQOO_FAIL_IF_QUIESCING;

	}
}
